/*! edge-builder v0.13.8 by Sebastian Software <s.werner@sebastian-software.de> */
import { getHashDigest } from 'loader-utils';
import _classCallCheck from 'babel-runtime/helpers/classCallCheck';
import ora from 'ora';
import { get } from 'app-root-dir';
import { relative, resolve } from 'path';
import fs, { existsSync } from 'fs';
import { readJsonSync, remove } from 'fs-extra';
import builtinModules from 'builtin-modules';
import resolvePkg from 'resolve-pkg';
import webpack from 'webpack';
import chalk from 'chalk';
import 'webpack/package.json';
import ExtractCssChunks from 'extract-css-chunks-webpack-plugin';
import StatsPlugin from 'stats-webpack-plugin';
import CaseSensitivePathsPlugin from 'case-sensitive-paths-webpack-plugin';
import SriPlugin from 'webpack-subresource-integrity';
import BabelMinifyPlugin from 'babel-minify-webpack-plugin';
import UglifyPlugin from 'uglifyjs-webpack-plugin';
import BundleAnalyzerPlugin from 'webpack-bundle-analyzer';
import { promisify } from 'bluebird';
import formatWebpackMessages from 'react-dev-utils/formatWebpackMessages';
import { notify } from 'edge-common';
import webpackDevMiddleware from 'webpack-dev-middleware';
import webpackHotMiddleware from 'webpack-hot-middleware';
import webpackHotServerMiddleware from 'webpack-hot-server-middleware';

function compareModules(left, right) {
  if (left.resource < right.resource) {
    return -1;
  }
  if (left.resource > right.resource) {
    return 1;
  }
  return 0;
}

function getModuleSource(module) {
  var source = module._source || {};
  return source._value || "";
}

function concatenateSource(result, moduleSource) {
  return result + moduleSource;
}

function ChunkHash() {

  // empty

}

var hashType = "sha256";
var digestType = "base62";
var digestLength = 8;

ChunkHash.prototype.apply = function (compiler) {
  compiler.plugin("compilation", function (compilation) {
    compilation.plugin("chunk-hash", function (chunk, chunkHash) {
      var source = chunk.mapModules(function (module) {
        return module;
      }).sort(compareModules).map(getModuleSource)

      // we provide an initialValue in case there is an empty module source. Ref: http://es5.github.io/#x15.4.4.21
      .reduce(concatenateSource, "");

      var generatedHash = getHashDigest(source, hashType, digestType, digestLength);
      chunkHash.digest = function () {
        return generatedHash;
      };
    });
  });
};

/*
	MIT License http://www.opensource.org/licenses/mit-license.php
	Author Tobias Koppers @sokra
*/

/* eslint-disable no-magic-numbers */

var ROOT = get();

var VerboseProgress = function () {
  function VerboseProgress(_ref) {
    var prefix = _ref.prefix;

    _classCallCheck(this, VerboseProgress);

    this.prefix = prefix;
  }

  VerboseProgress.prototype.apply = function apply(compiler) {
    var spinner = null;
    var lastModule = null;
    var prefix = this.prefix ? this.prefix + " " : "";

    function display(message) {
      if (message !== "") {
        spinner.text = prefix + message;

        // We somehow have to force rendering otherwise busy Webpack wouldn't let us.
        spinner.render();
      } else {
        spinner.succeed(prefix + "Done!");
      }
    }

    function moduleDone() {

      var humanIndex = void 0;
      var humanModuleId = lastModule;

      humanIndex = humanModuleId.lastIndexOf(" ");
      humanModuleId = humanIndex === -1 ? humanModuleId : humanModuleId.slice(humanIndex + 1, humanModuleId.length);

      humanIndex = humanModuleId.lastIndexOf("!");
      humanModuleId = humanIndex === -1 ? humanModuleId : humanModuleId.slice(humanIndex + 1, humanModuleId.length);

      humanIndex = humanModuleId.indexOf("?");
      humanModuleId = humanIndex === -1 ? humanModuleId : humanModuleId.slice(0, humanIndex);

      humanModuleId = relative(ROOT, humanModuleId).replace(/^node_modules\//, "~/");

      if (humanModuleId.startsWith('"') && humanModuleId.endsWith('"')) {
        humanModuleId = humanModuleId.slice(1, -1);
      }

      // Ignore Context Logic Imports
      if (humanModuleId.includes("|")) {
        return;
      }

      // Exclude hard to read directly relative modules
      if (humanModuleId.startsWith("..")) {
        return;
      }

      if (/^[a-zA-Z0-9\-_/~\.]{0,50}$/.test(humanModuleId)) {
        display("Building Modules " + humanModuleId + "...");
      }
    }

    compiler.plugin("compilation", function (compilation) {
      if (compilation.compiler.isChild()) {
        return;
      }

      spinner = ora({ interval: 16 });
      spinner.start();

      display(0, "compiling");

      compilation.plugin("build-module", function (module) {
        lastModule = module.identifier();
      });

      compilation.plugin("failed-module", moduleDone);
      compilation.plugin("succeed-module", moduleDone);

      var syncHooks = {
        "seal": "Sealing",
        "optimize": "Optimizing",
        "optimize-modules-basic": "Optimizing modules",
        "optimize-chunks-basic": "Optimizing chunks",
        "optimize-chunk-modules": "Optimizing chunk modules",
        "optimize-module-order": "Optimizing module order",
        "optimize-module-ids": "Optimizing module ids",
        "optimize-chunk-order": "Optimizing chunk order",
        "optimize-chunk-ids": "Optimizing chunk ids",
        "before-hash": "Hashing",
        "before-module-assets": "Processing module assets",
        "before-chunk-assets": "Processing chunk assets",
        "record": "Recording"
      };

      Object.keys(syncHooks).forEach(function (name) {
        var pass = 0;
        var message = syncHooks[name];
        compilation.plugin(name, function () {
          if (pass++ > 0) display(message + (" [pass " + pass + "]"));else display(message);
        });
      });

      compilation.plugin("optimize-tree", function (chunks, modules, callback) {
        display("Optimizing tree");
        callback();
      });

      compilation.plugin("additional-assets", function (callback) {
        display("Processing assets");
        callback();
      });

      compilation.plugin("optimize-chunk-assets", function (chunks, callback) {
        display("Optimizing chunk assets");
        callback();
      });

      compilation.plugin("optimize-assets", function (assets, callback) {
        display("Optimizing assets");
        callback();
      });
    });

    compiler.plugin("emit", function (compilation, callback) {
      display("Emitting");
      callback();
    });

    compiler.plugin("done", function () {
      display("");
    });
  };

  return VerboseProgress;
}();

var BuiltIns = new Set(builtinModules);

var BundleWithWebpack = new Set([
// These are required for universal import handling.
"react-universal-component", "webpack-flush-chunks", "babel-plugin-universal-import"]);

var Problematic = new Set([
// "intl" is included in one block with complete data of all locales. There is no benefit for bundle everything here.
"intl",

// "mime-db" database for working with mime types. Naturally pretty large stuff.
"mime-db",

// "encoding" uses dynamic iconv loading
"encoding",

// These packages make use of dynamic require expressions which do not work well with Webpack
"ajv", "colors", "jsdom"]);

function isLoaderSpecificFile(request) {
  // Webpack Loader Syntax
  if (request.charAt(0) === "!") {
    return true;
  }

  return Boolean(
  // eslint-disable-next-line max-len
  /\.(eot|woff|woff2|ttf|otf|svg|png|jpg|jpeg|gif|webp|webm|ico|mp4|mp3|ogg|html|pdf|css|scss|sass|sss|less|zip)$/.exec(request));
}

var bundleCache = {};

function shouldPackageBeBundled(resolved) {
  var result = null;

  // Detect Node-Gyp Bindings File
  // "describes the configuration to build your module in a JSON-like format"
  // Via: https://github.com/nodejs/node-gyp#the-bindinggyp-file
  var hasBindings = existsSync(resolve(resolved, "bindings.gyp"));
  if (hasBindings) {
    result = false;
  } else {
    var json;

    try {
      json = readJsonSync(resolve(resolved, "package.json"));
    } catch (error) {
      // pass
    }

    if (json) {
      if (json.module || json.style || json.browser || json["jsnext:main"]) {
        result = true;
      }

      // Configuration for Node-Pre-Gyp
      // See also: https://www.npmjs.com/package/node-pre-gyp
      if (json.binary != null) {
        result = false;
      }
    }
  }

  return result;
}

function shouldBeBundled(basename) {
  if (basename in bundleCache) {
    return bundleCache[basename];
  }

  var resolved = void 0;
  try {
    resolved = resolvePkg(basename);
  } catch (error) {
    return null;
  }

  var result = resolved ? shouldPackageBeBundled(resolved) : null;
  bundleCache[basename] = result;
  return result;
}

// eslint-disable-next-line
function isRequestExternal(request) {
  var lightweight = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : false;

  // Ignore all inline files for further processing
  if (request.charAt(0) === ".") {
    return false;
  }

  // Inline all files which are dependend on Webpack loaders e.g. CSS, images, etc.
  if (isLoaderSpecificFile(request)) {
    return false;
  }

  var match = /^((@[a-zA-Z0-9-_]+\/)?[a-zA-Z0-9_-]+)\/?/.exec(request);
  var basename = match ? match[1] : null;

  if (basename == null) {
    console.warn("Couldn't figure out package name from require:", request);

    // We assume that all special scenarios which are not covered yet, are
    // probably some extension in Webpack and should therefor be bundled.
    return true;
  }

  // Externalize built-in modules
  if (BuiltIns.has(basename)) {
    return true;
  }

  // Inline all modules which require a Webpack environment
  if (BundleWithWebpack.has(basename)) {
    return false;
  }

  // Make sure that problematic CommonJS code is externalized
  if (Problematic.has(basename)) {
    return true;
  }

  // Analyses remaining packages whether these offer ES2015 bundles
  // and/or include native extensions via Gyp. We try to bundle all
  // ES2015 modules as it is better for tree-shaking. We can't bundle
  // any modules which contain binary code. If neither of these two is
  // true we just follow the wishes of the user regarding bundling
  // lightweight or heavyweight to include the maximum amount or least
  // amount of modules feasible.
  var bundle = shouldBeBundled(basename);
  if (bundle != null) {
    return !bundle;
  }

  // Finally we respect the user given preference
  return lightweight;
}

var externalsCache = {};

function getServerExternals(lightweight, entries) {
  var entriesSet = new Set(entries);

  // We can't influence a public interface
  // eslint-disable-next-line max-params
  return function (context, request, callback) {
    // Make sure to include all direct entries
    if (entriesSet.has(request)) {
      return callback();
    }

    var isExternal = externalsCache[request];
    if (isExternal == null) {
      isExternal = isRequestExternal(request, lightweight);
      externalsCache[request] = isExternal;
    }

    return isExternal ? callback(null, "commonjs " + request) : callback();
  };
}

/* eslint-disable max-params, max-statements, complexity */

// Using more modern approach of hashing than "webpack-md5-hash". Somehow the SHA256 version
// ("webpack-sha-hash") does not correctly work based (produces different hashes for same content).
// This is basically a replacement of md5 with the loader-utils implementation which also supports
// shorter generated hashes based on base62 encoding instead of hex.
// CSS Support
// Core
// Generating static HTML pages
// import HtmlWebpackPlugin from "html-webpack-plugin"
// Compression
// import ZopfliPlugin from "zopfli-webpack-plugin"

function removeEmptyKeys(source) {
  var copy = {};
  for (var key in source) {
    if (!(source[key] == null || source[key].length === 0)) copy[key] = source[key];
  }

  return copy;
}

// https://github.com/mishoo/UglifyJS2#compress-options
var UGLIFY_OPTIONS = {
  /* eslint-disable camelcase */

  compress: {
    // Only risky for some rare floating point situations
    unsafe_math: true,

    // optimize expressions like Array.prototype.slice.call(a) into [].slice.call(a)
    unsafe_proto: true,

    // Good for Chrome performance
    keep_infinity: true,

    // Try harder to export less code
    passes: 2
  },

  output: {
    // Fix for problematic code like emoticons
    ascii_only: true,

    // More readable output
    // Whenever possible we will use a newline instead of a semicolon
    // semicolons: false,

    // Remove all comments, don't even keep tons of copyright comments
    comments: false
  }
};

var BABEL_MINIFY_CLIENT_OPTIONS = {};

var BABEL_SERVER_MINIFY_OPTIONS = {
  booleans: false,
  deadcode: true,
  flipComparisons: false,
  mangle: false,
  mergeVars: false
};

var ROOT$1 = get();

var assetFiles = /\.(eot|woff|woff2|ttf|otf|svg|png|jpg|jpeg|jp2|jpx|jxr|gif|webp|mp4|mp3|ogg|pdf|html|ico)$/;
var babelFiles = /\.(js|mjs|jsx)$/;
var postcssFiles = /\.(css|sss|pcss)$/;
var identityFnt = function (item) {
  return item;
};

function builder(target) {
  var env = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : "development";
  var config = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};

  var SERVER_OUTPUT = config.output.server;
  var CLIENT_OUTPUT = config.output.client;

  // const HTML_TEMPLATE = config.entry.htmlTemplate
  var BABEL_ENV = config.build.babelEnvPrefix + "-" + env + "-" + target;

  var PROJECT_CONFIG = require(resolve(ROOT$1, "package.json"));

  var CACHE_HASH = getHashDigest(JSON.stringify(PROJECT_CONFIG), "sha256", "base62", 4);

  var PREFIX = chalk.bold(target.toUpperCase());

  var DEFAULT_LOCALE = config.locale["default"];
  var SUPPORTED_LOCALES = config.locale.supported;

  var isServer = target === "server";
  var isClient = target === "client";

  var isDevelopment = env === "development";
  var isProduction = env === "production";

  // Extract plain languages from configured locales
  var SUPPORTED_LANGUAGES = function () {
    var languages = new Set();
    for (var _iterator = SUPPORTED_LOCALES, _isArray = Array.isArray(_iterator), _i = 0, _iterator = _isArray ? _iterator : _iterator[Symbol.iterator]();;) {
      var _ref;

      if (_isArray) {
        if (_i >= _iterator.length) break;
        _ref = _iterator[_i++];
      } else {
        _i = _iterator.next();
        if (_i.done) break;
        _ref = _i.value;
      }

      var entry = _ref;

      languages.add(entry.split("-")[0]);
    }
    return Array.from(languages.keys());
  }();

  var LEAN_INTL_REGEXP = new RegExp("\\b" + SUPPORTED_LOCALES.join("\\b|\\b") + "\\b");
  var REACT_INTL_REGEXP = new RegExp("\\b" + SUPPORTED_LANGUAGES.join("\\b|\\b") + "\\b");

  var name = isServer ? "server" : "client";
  var webpackTarget = isServer ? "node" : "web";
  var devtool = config.build.enableSourceMaps ? "source-map" : false;

  console.log(chalk.underline(PREFIX + " Configuration:"));
  console.log("\u2192 Environment: " + env);
  console.log("\u2192 Webpack Target: " + webpackTarget);

  if (config.verbose) {
    console.log("\u2192 Babel Environment: " + BABEL_ENV);
    console.log("\u2192 Enable Source Maps: " + devtool);
    console.log("\u2192 Bundle Compression: " + config.build.bundleCompression);
    console.log("\u2192 Use Cache Loader: " + config.build.useCacheLoader + " [Hash: " + CACHE_HASH + "]");
    console.log("\u2192 Default Locale: " + DEFAULT_LOCALE);
    console.log("\u2192 Supported Locales: " + SUPPORTED_LOCALES);
    console.log("\u2192 Supported Languages: " + SUPPORTED_LANGUAGES);
  }

  var CACHE_LOADER_DIRECTORY = resolve(ROOT$1, ".cache/loader-" + CACHE_HASH + "-" + target + "-" + env);

  var cacheLoader = config.build.useCacheLoader ? {
    loader: "cache-loader",
    options: {
      cacheDirectory: CACHE_LOADER_DIRECTORY
    }
  } : null;

  var cssLoaderOptions = {
    modules: true,
    localIdentName: "[local]-[hash:base62:8]",
    "import": false,

    // We are using CSS-O as part of our PostCSS-Chain
    minimize: false,
    sourceMap: config.build.enableSourceMaps
  };

  var postCSSLoaderRule = {
    loader: "postcss-loader",
    query: {
      sourceMap: config.build.enableSourceMaps
    }

    // Just bundle the NodeJS files which are from the local project instead
    // of a deep self-contained bundle.
    // See also: https://nolanlawson.com/2016/08/15/the-cost-of-small-modules/
  };

  var HMR_MIDDLEWARE = "webpack-hot-middleware/client?path=/__webpack_hmr&timeout=10000&reload=true&noInfo=true&overlay=false";

  var VENDOR_ENTRY = isServer ? config.entry.serverVendor : config.entry.clientVendor;
  var MAIN_ENTRY = isServer ? config.entry.serverMain : config.entry.clientMain;

  var HAS_VENDOR = fs.existsSync(VENDOR_ENTRY);
  var HAS_MAIN = fs.existsSync(MAIN_ENTRY);

  var WEBPACK_HOOK = config.hook.webpack ? config.hook.webpack : identityFnt;

  return WEBPACK_HOOK({
    name: name,
    target: webpackTarget,
    devtool: devtool,
    context: ROOT$1,
    externals: isServer ? getServerExternals(isDevelopment, [VENDOR_ENTRY, MAIN_ENTRY]) : undefined,

    entry: removeEmptyKeys({
      vendor: HAS_VENDOR ? [HAS_VENDOR && isClient && isDevelopment ? HMR_MIDDLEWARE : null, VENDOR_ENTRY].filter(Boolean) : null,
      main: HAS_MAIN ? [HAS_MAIN && isClient && isDevelopment ? HMR_MIDDLEWARE : null, MAIN_ENTRY].filter(Boolean) : null
    }),

    output: {
      libraryTarget: isServer ? "commonjs2" : "var",
      filename: isDevelopment || isServer ? "[name].js" : "[name]-[chunkhash].js",
      chunkFilename: isDevelopment || isServer ? "[name].js" : "[name]-[chunkhash].js",
      path: isServer ? SERVER_OUTPUT : CLIENT_OUTPUT,

      // This is required for loading dynamic chunks and other files. Should point
      // to the CDN URL if that is used.
      publicPath: config.output["public"],

      // Tell webpack to include comments in bundles with information about the
      // contained modules. This option defaults to false and should not be used
      // in production, but it's very useful in development when reading the generated code.
      // https://webpack.js.org/configuration/output/#output-pathinfo
      pathinfo: isDevelopment,

      // Enable cross-origin loading without credentials - Useful for loading files from CDN
      // when React server is hosted on a different domain.
      crossOriginLoading: "anonymous"
    },

    module: {
      rules: [{
        test: babelFiles,
        loader: "source-map-loader",
        enforce: "pre",
        options: {
          quiet: true
        },
        exclude: [
        // These packages point to sources which do not exist
        // See also: https://github.com/webpack-contrib/source-map-loader/issues/18
        // Waiting for PR: https://github.com/webpack-contrib/source-map-loader/pull/50
        /intl-/, /apollo-/, /react-apollo/, /zen-observable-ts/]
      },

      // References to images, fonts, movies, music, etc.
      {
        test: assetFiles,
        loader: "file-loader",
        options: {
          name: isProduction ? "file-[hash:base62:8].[ext]" : "[path][name].[ext]",
          emitFile: isClient
        }
      },

      // YAML
      {
        test: /\.(yml|yaml)$/,
        loaders: ["json-loader", "yaml-loader"]
      },

      // GraphQL support
      // @see http://dev.apollodata.com/react/webpack.html
      {
        test: /\.(graphql|gql)$/,
        loader: "graphql-tag/loader"
      },

      // Transpile our own JavaScript files using the setup in `.babelrc`.
      {
        test: babelFiles,
        exclude: /node_modules/,
        use: [
        // Note:
        // We prefer cache-loader over babel cache mechanism. Reason:
        // "They both serve the same purpose and are interchangeable, with cache-loader being the 'newer and official' way to cache a loader in general"
        cacheLoader, {
          loader: "babel-loader",
          options: {
            babelrc: true,
            forceEnv: BABEL_ENV
          }
        }].filter(Boolean)
      },

      // Use either
      {
        test: postcssFiles,
        use: isClient ? ExtractCssChunks.extract({
          use: [cacheLoader, {
            loader: "css-loader",
            options: cssLoaderOptions
          }, postCSSLoaderRule].filter(Boolean)
        }) : [cacheLoader, {
          loader: "css-loader/locals",
          options: cssLoaderOptions
        }, postCSSLoaderRule].filter(Boolean)
      }].filter(Boolean)
    },

    plugins: [
    // Completely filter out locale data for locales we definitely do not support
    // This is actually massive for bundling times, bundle sizes and all.
    // Currently supports lean-intl and react-intl. We should propbably add more libraries
    // here over time e.g. momentjs and other locale data dependent stuff.
    new webpack.ContextReplacementPlugin(/lean-intl\/locale-data/, LEAN_INTL_REGEXP), new webpack.ContextReplacementPlugin(/react-intl\/locale-data/, REACT_INTL_REGEXP), new webpack.DefinePlugin({
      // Important:
      // We keep all these env-variables separate as this allows for
      // parallel usage of libraries like `dotenv` do make non compile-time
      // environment variables accessible.
      "process.env.NODE_ENV": JSON.stringify(env),
      "process.env.BUILD_TARGET": JSON.stringify(webpackTarget)
    }),

    // Generating static HTML page for simple static deployment
    // https://github.com/jantimon/html-webpack-plugin
    /*
    isProduction && isClient ?
      new HtmlWebpackPlugin({
        template: HTML_TEMPLATE
      }) :
      null,
    */

    // Subresource Integrity (SRI) is a security feature that enables browsers to verify that
    // files they fetch (for example, from a CDN) are delivered without unexpected manipulation.
    // https://www.npmjs.com/package/webpack-subresource-integrity
    // Browser-Support: http://caniuse.com/#feat=subresource-integrity
    new SriPlugin({
      hashFuncNames: ["sha256", "sha512"],
      enabled: isProduction && isClient
    }),

    // Improve OS compatibility
    // https://github.com/Urthen/case-sensitive-paths-webpack-plugin
    new CaseSensitivePathsPlugin(),

    // Custom progress plugin
    process.stdout.isTTY ? new VerboseProgress({
      prefix: PREFIX
    }) : null,

    // Analyse bundle in production
    isClient && isProduction ? new BundleAnalyzerPlugin.BundleAnalyzerPlugin({
      analyzerMode: "static",
      defaultSizes: "gzip",
      logLevel: "silent",
      openAnalyzer: false,
      reportFilename: "report.html"
    }) : null,

    // Analyse bundle in production
    isServer && isProduction ? new BundleAnalyzerPlugin.BundleAnalyzerPlugin({
      analyzerMode: "static",
      defaultSizes: "parsed",
      logLevel: "silent",
      openAnalyzer: false,
      reportFilename: "report.html"
    }) : null,

    // We use this so that our generated [chunkhash]'s are only different if
    // the content for our respective chunks have changed.  This optimises
    // our long term browser caching strategy for our client bundle, avoiding
    // cases where browsers end up having to download all the client chunks
    // even though 1 or 2 may have only changed.
    isProduction && isClient ? new ChunkHash() : null,

    // Let the server side renderer know about our client side assets
    // https://github.com/FormidableLabs/webpack-stats-plugin
    isProduction && isClient ? new StatsPlugin("stats.json") : null,

    // Classic UglifyJS for compressing ES5 compatible code.
    // https://github.com/webpack-contrib/uglifyjs-webpack-plugin
    config.build.bundleCompression === "uglify" && isProduction && isClient ? new UglifyPlugin({
      sourceMap: config.build.enableSourceMaps,
      cache: true,
      parallel: true,
      uglifyOptions: UGLIFY_OPTIONS
    }) : null,

    // Alternative to Uglify when producing modern output
    // Advanced ES2015 ready JS compression based on Babylon (Babel Parser)
    // https://github.com/webpack-contrib/babili-webpack-plugin
    config.build.bundleCompression === "babel" && isProduction && isClient ? new BabelMinifyPlugin(BABEL_MINIFY_CLIENT_OPTIONS, { comments: false }) : null, isProduction && isServer ? new BabelMinifyPlugin(BABEL_SERVER_MINIFY_OPTIONS, { comments: false }) : null,

    // "Use HashedModuleIdsPlugin to generate IDs that preserves over builds."
    // Via: https://github.com/webpack/webpack.js.org/issues/652#issuecomment-273324529
    isProduction ? new webpack.HashedModuleIdsPlugin() : null,

    // I would recommend using NamedModulesPlugin during development (better output).
    // Via: https://github.com/webpack/webpack.js.org/issues/652#issuecomment-273023082
    isDevelopment ? new webpack.NamedModulesPlugin() : null, isClient ? new ExtractCssChunks({
      filename: isDevelopment ? "[name].css" : "[name]-[contenthash:base62:8].css"
    }) : null, isServer ? new webpack.optimize.LimitChunkCountPlugin({ maxChunks: 1 }) : null,

    // Extract Webpack bootstrap code with knowledge about chunks into separate cachable package.
    // isClient ? new webpack.optimize.CommonsChunkPlugin({
    //   names: [ "bootstrap" ],

    //   // needed to put webpack bootstrap code before chunks
    //   filename: isProduction ? "[name]-[chunkhash].js" : "[name].js",
    //   minChunks: Infinity
    // }) : null,

    // https://webpack.js.org/plugins/commons-chunk-plugin/#explicit-vendor-chunk
    HAS_VENDOR && isClient ? new webpack.optimize.CommonsChunkPlugin({
      name: "vendor",

      // With more entries, this ensures that no other module goes into the vendor chunk
      minChunks: Infinity
    }) : null, isProduction ? new webpack.optimize.ModuleConcatenationPlugin() : null, isClient && isDevelopment ? new webpack.HotModuleReplacementPlugin() : null, isDevelopment ? new webpack.NoEmitOnErrorsPlugin() : null].filter(Boolean)
  }, {
    isServer: isServer,
    isClient: isClient,
    isProduction: isProduction,
    isDevelopment: isDevelopment
  });
}

var removePromise = promisify(remove);

/* eslint-disable no-console, max-params */
function getWebpackHandler(target, resolve$$1, reject) {
  return function (fatalError, stats) {
    if (fatalError) {
      notify("Fatal error during compiling " + target + ": " + fatalError, "error");
      return reject();
    }

    var rawMessages = stats.toJson({});
    var messages = formatWebpackMessages(rawMessages);

    var isSuccessful = !messages.errors.length && !messages.warnings.length;
    if (isSuccessful) {
      notify("Compiled " + target + " successfully!", "info");
    }

    // If errors exist, only show errors.
    if (messages.errors.length) {
      notify("Failed to compile " + target + "!", "error");
      console.log(messages.errors.join("\n\n"));
      return reject();
    }

    return resolve$$1();
  };
}

function buildClient() {
  var config = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};

  var webpackConfig = builder("client", "production", config);

  return new Promise(function (resolve$$1, reject) {
    webpack(webpackConfig, getWebpackHandler("client", resolve$$1, reject));
  });
}

function buildServer() {
  var config = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};

  var webpackConfig = builder("server", "production", config);

  return new Promise(function (resolve$$1, reject) {
    webpack(webpackConfig, getWebpackHandler("server", resolve$$1, reject));
  });
}

function cleanServer() {
  var config = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};

  return removePromise(config.output.server);
}

function cleanClient() {
  var config = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};

  return removePromise(config.output.client);
}

function createMiddleware() {
  var config = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};

  var clientConfig = builder("client", "development", config);
  var serverConfig = builder("server", "development", config);

  var multiCompiler = webpack([clientConfig, serverConfig]);
  var clientCompiler = multiCompiler.compilers[0];

  var devMiddleware = webpackDevMiddleware(multiCompiler, {
    // required
    publicPath: config.output["public"],

    // we have our custom error handling for webpack which offers far better DX
    quiet: true,

    // display no info to console (only warnings and errors)
    noInfo: true
  });

  var hotMiddleware = webpackHotMiddleware(clientCompiler);

  // keeps serverRender updated with arg: { clientStats, outputPath }
  var hotServerMiddleware = webpackHotServerMiddleware(multiCompiler, {
    serverRendererOptions: {
      outputPath: config.output.client
    }
  });

  return {
    middleware: [devMiddleware, hotMiddleware, hotServerMiddleware],
    multiCompiler: multiCompiler
  };
}

function connectWithWebpack(server, multiCompiler) {
  var serverIsStarted = false;

  multiCompiler.plugin("invalid", function () {
    notify("Compiling...", "info");
  });

  multiCompiler.plugin("done", function (stats) {
    /* eslint-disable no-console */
    var rawMessages = stats.toJson({});
    var messages = formatWebpackMessages(rawMessages);

    var isSuccessful = !messages.errors.length && !messages.warnings.length;
    if (isSuccessful) {
      notify("Compiled successfully!", "info");
    }

    // If errors exist, only show errors.
    if (messages.errors.length) {
      notify("Failed to compile!", "error");
      console.log(messages.errors.join("\n\n"));
      return;
    }

    // Show warnings if no errors were found.
    if (messages.warnings.length) {
      notify("Compiled with warnings!", "warn");
      console.log(messages.warnings.join("\n\n"));
    }

    if (!stats.hasErrors() && !serverIsStarted) {
      serverIsStarted = true;

      server.listen(process.env.SERVER_PORT, function () {
        notify("Server started at port " + process.env.SERVER_PORT, "info");
      });
    }
  });
}

export { cleanClient, cleanServer, buildClient, buildServer, connectWithWebpack, createMiddleware };
//# sourceMappingURL=node.esm.js.map
